//	Author: Patrick Reilly
//	ID:		70274,161 (Compuserve)
//	Version	1.01

//	Contents --------------------------------------------------------------
//
//      termbuf::termbuf						constructor
//		termbuf::~termbuf						destructor
//		termbuf::setbuf
//      termbuf::underflow
//      termbuf::overflow
//		termbuf::setmap
//		termbuf::setWindow
//		termbuf::setBackground
//		termbuf::setForeground
//		termbuf::setAttribute
//		termbuf::clearWindow
//		termbuf::gotoWindow
//      termbuf::initScreen
//      termbuf::processKey
//
//  Description
//
//      Class termbuf is derived from streambuf for use in terminal
//	emulation (for telecommunication apps). The key to emulation is
//	mapping keystrokes (scan+ascii) to both actions and terminal
//	emulation strings. termbuf class does this with the typedef structure
//	KeyMap. Certain keystrokes are hardcoded to certain actions
//	(ie ascii 30h prints the letter '0' and ^PGDN clears the screen). In
//	the KeyMap structure they are then cross-referenced to the emulation
//	string for that keycode. For example, the keycode 0x5000 (DOWN key) is
//	hardcoded to move the cursor down one line, and (in VT102 emulation)
//	softcoded to "\x01B[B". Therefore if you extract the keystroke 'down-
//	arrow' you will extract the string "\x01B[B", and if you insert the
//	string "\x01B[B" the cursor will move down the screen one row.
//
//	End ------------------------------------------------------------------

//	Interface Dependencies -----------------------------------------------

#ifndef __TERM_H
#include "term.h"
#endif

#ifndef __BIOS_H
#include <bios.h>
#endif

#ifndef __CTYPE_H
#include <ctype.h>
#endif

#ifndef __STRING_H
#include <string.h>
#endif

//	End Interface Dependencies --------------------------------------------

//	Constructor //

_Cdecl termbuf::termbuf()

//	Summary --------------------------------------------------------------
//
//		Constructor for class termbuf.
//
//	Functional Description
//
//		The two constructors for termbuf are identical. It sets the base
//	streambuf class buffer pointers to NULL so that all incoming and out-
//	going characters will be forced to call member underflow() and
//	overflow(). It sets the key mapping pointers to NULL to indicate that
//	no emulation is in effect. It initializes emapbuf to the end of the
//	input buffer (mapbuf)+1 and the input put pointer (pmap) to the start
//	of the buffer. It fills the mapMatch array with ones to indicate that
//	no elements of the KeyMap array have been eliminated by earlier
//	entries. It initializes the text_info structure to the current screen
//	window, cursor location, and attribute.
//
//	End -------------------------------------------------------------------

	{
	setb(0,0);		setp(0,0);		setg(0,0,0);	// clear pointers.
	map = emap = 0;
	insLen = keymapsize = 0;
	gettextinfo(&ti);
	}

//	End Constructor //

//	Constructor //

_Cdecl termbuf::termbuf(char _FAR *,int)

//	Summary --------------------------------------------------------------
//
//		Constructor for class termbuf.
//
//	Functional Description
//
//		The two constructors for termbuf are identical. It sets the base
//	streambuf class buffer pointers to NULL so that all incoming and out-
//	going characters will be forced to call member underflow() and
//	overflow(). It sets the key mapping pointers to NULL to indicate that
//	no emulation is in effect. It initializes emapbuf to the end of the
//	input buffer (mapbuf)+1 and the input put pointer (pmap) to the start
//	of the buffer. It fills the mapMatch array with ones to indicate that
//	no elements of the KeyMap array have been eliminated by earlier
//	entries. It initializes the text_info structure to the current screen
//	window, cursor location, and attribute.
//
//	End -------------------------------------------------------------------

	{
	setb(0,0);		setp(0,0);		setg(0,0,0);
	map = emap = 0;
	insLen = keymapsize = 0;
	gettextinfo(&ti);
	}

//	End Constructor //

//	Destructor //
_Cdecl termbuf::~termbuf()

//	Summary --------------------------------------------------------------
//
//		Vanilla destructor.
//
//	End ------------------------------------------------------------------

	{
	}

//	Member Function //

streambuf _FAR * _Cdecl termbuf::setbuf(signed char _FAR *,int)

//	Summary --------------------------------------------------------------
//
//		Overloaded base member.
//
//	Functional Description
//
//		This member is overloaded because a streambuf buffer is not
//	maintained in the conventional manner.
//
//	End ------------------------------------------------------------------

	{
	return this;
	}

//	End Member Function //

//	Member Function //

int _Cdecl termbuf::underflow()

//	Summary --------------------------------------------------------------
//
//		Overloaded member called when extraction fails.
//
//	Functional Description
//
//      Extraction takes place under one of two conditions: 1) the streambuf
//	'get' buffer is empty and you need to get a character from the keyboard,
//	or 2) you previously extracted from the keyboard an are getting
//	subsequent chars from a multi-character emulation (ie one keystroke
//	equates to 2+ emulation chars). If underflow() is called with the get
//	buffer empty, it will call bioskey(0) to wait for a keystroke response
//	from the user. It then searches the KeyMap array for a matching entry.
//	If one is not found, it loops back for another key. If a match is found,
//	the streambuf get pointers are set to the emulation string that
//	corresponds to that keystroke. This way, subsequent extraction will
//	read all of the emulation string before underflow() is called again.
//		If the KeyMap pointer is NULL (no emulation), each keystroke is
//	returned with no translation.
//
//	End ------------------------------------------------------------------

	{
	char *p,*e;
	int x,lo,hi,i;

	if(!map)			// no emulation
		{
		setb(&noMap,&noMap+1);
		setg(&noMap,&noMap,&noMap+1);
		noMap = bioskey(0) & 0xFF;
		return (int)noMap;
		}

	while(1)
		{
		x = bioskey(0);

		// perform binary search for keycode //
		for(lo = 0,hi = keymapsize-1; lo != hi;)
			{
			i = lo+(hi-lo)/2;
			if(map[i].keycode == x)	break;
			else if(map[i].keycode < x)	lo = i;
			else						hi = i;
			}
		if(map[i].keycode == x)	break;
		}
	p = map[i].string;
	for(e = p; *e; e++);

	// set the get and buffer pointers to the emulation string //
	setb(p,e);
	setg(p,p,e);
	return *p;
	}

//	End Member Function //

//	Member Function //

int _Cdecl termbuf::overflow(int ch)

//	Summary --------------------------------------------------------------
//
//		Overloaded insertion member.
//
//	Functional Description
//
//      Since the streambuf 'put' pointers are maintained at NULL, all
//	insertion will cause overflow() to be called for each character. There
//	are two cases overflow() must handle: emulation is in effect (the KeyMap
//	pointer member map is not NULL) or it is not in effect.
//		Class termbuf maintains a member mapMatch, which is just an array
//	of flags that correspond to each KeyMap element. When a character is
//	inserted, overflow() searches all of the KeyMap emulation strings for
//	a match. If an element does not match, the corresponding mapMatch flag
//	is set to false. If a match IS found and the next character in the
//	emulation string is a NULL (end of string), the emulation has been
//	completed and processKey() member is called for the corresponding
//	keycode and the member returns. If there was no match in any of the
//	emulation strings (all elements of mapMatch false), then the insertion
//	was unrecognizable and everything is reinitialized.
//		If no emulation is in effect (KeyMap pointer member map is NULL), the
//	character inserted is sent to processKey().
//
//	NOTE:
//
//		Please note that insertion emulation is performed as follows: an
//	inserted string is correlated to a keycode in the KeyMap array; the
//	keycode is sent to processKey(), which hard-codes certain keycodes to
//	certain actions. Because the KeyMap array is searched sequentially,
//	you cannot have a lower-element emulation string be a sub-set of a
//	higher-element emulation string and have the second string be executed
//	(ie if element map[n].string = "xyz" and map[n+m].string = "xyzabc" then
//	map[n+m].keycode will NEVER be sent to processKey() because the input
//	will match "xyz", execute, reinitialize, and then "abc" will be
//	processed.
//
//	End ------------------------------------------------------------------

	{
	int i;

	// don't process EOF entry //
	if(ch == EOF)	return ch;

	// no emulation //
	if(!map)
		{
		initScreen();
		processKey(ch);
		gettextinfo(&ti);
		return ch;
		}

	// search map for entries that match //
	for(i = 0; i < keymapsize; i++)
		{
		// if non-matching entry, continue //
		if(!mapMatch[i])	continue;

		// process no match //
		if(map[i].string[insLen] != ch)
			{
			mapMatch[i] = 0;
			continue;
			}

		// process for completed emulation //
		if(!map[i].string[insLen+1])
			{
			initScreen();
			processKey(map[i].keycode);
			gettextinfo(&ti);
			insLen = 0;
			memset(mapMatch,1,keymapsize);
			return ch;
			}
		}

	// check for unrecognized input //
	if(!memchr(mapMatch,1,keymapsize))
		{
		insLen = 0;
		memset(mapMatch,1,keymapsize);
		}
	insLen++;
	return ch;
	}

//	End Member Function //

//	Member Function //

void _Cdecl termbuf::setmap(KeyMap _FAR *p)

//	Summary ---------------------------------------------------------------
//
//		Initialize the key mapping array KeyMap.
//
//	Functional Description
//
//		The key mapping KeyMap array is an array of KEYMAPSIZE that is used
//	for keycode-to-emulation mapping. Member setmap is used to initialize
//	the array pointers to install a particular terminal emulation.
//
//	End -------------------------------------------------------------------

	{
	map = p;
	if(map)
		{
		for(keymapsize = 0; map[keymapsize].keycode; keymapsize++);
		emap = map+keymapsize;
	else
		{
		keymapsize = 0;
		emap = map;
		}

	// initialize so that any current emulation in progress is aborted //
	insLen = 0;
	memset(mapMatch,1,keymapsize);
	}

//	End Member Function //

//	Member Function //

void _Cdecl termbuf::setWindow(Rect& rect)

//	Summary -------------------------------------------------------------
//
//      Sets the screen coordinates for the terminal window.
//
//	End -----------------------------------------------------------------

	{
	setWindow(rect.left,rect.top,rect.right,rect.bottom);
	}

//	End Member Function //

//	Member Function //

void _Cdecl termbuf::setWindow(int left,int top,int right,int bottom)

//	Summary ------------------------------------------------------------
//
//      Sets the screen coordinates for the terminal window.
//
//	End ----------------------------------------------------------------

	{
	int i;

	if(left > right)	{	i = left;	left = right;	right = i;	}
	if(top > bottom)	{	i = top;	top = bottom;	bottom = i;	}
	if(left < 1)						left = 1;
	else if(left > ti.screenwidth)      return;
	if(right < 1)						return;
	else if(right > ti.screenwidth)		right = ti.screenwidth;
	if(top < 1)							top = 1;
	else if(top > ti.screenheight)		return;
	if(bottom < 1)						return;
	else if(bottom > ti.screenheight)	bottom = ti.screenheight;
	if(left >= right || top >= bottom)	return;
	initScreen();
	clrscr();
	ti.winleft = left;		ti.winright = right;
	ti.wintop = top;		ti.winbottom = bottom;
	ti.curx = ti.cury = 1;
	initScreen();
	}

//	End Member Function //

//	Member Function //

void _Cdecl termbuf::setBackground(int clr)

//	Summary -------------------------------------------------------------
//
//		Sets the terminal background color.
//
//	End -----------------------------------------------------------------

	{
	if(clr > 15)	return;
	clr <<= 4;
	ti.attribute = (ti.attribute & 0x0F) | clr;
	ti.normattr = (ti.normattr & 0x0F) | clr;
	}

//	End Member Function //

//	Member Function //

void _Cdecl termbuf::setForeground(int clr)

//	Summary -------------------------------------------------------------
//
//		Sets the terminal foreground color.
//
//	End -----------------------------------------------------------------

	{
	if(clr > 15)	return;
	ti.attribute = (ti.attribute & 0xFFF0) | clr;
	ti.normattr = (ti.normattr & 0xFFF0) | clr;
	}

//	End Member Function //

//	Member Function //

void _Cdecl termbuf::setAttribute(int attr)

//	Summary -------------------------------------------------------------
//
//		Sets the terminal foreground and background colors.
//
//	End -----------------------------------------------------------------

	{
	if(attr > 0xFF)	return;
	ti.attribute = ti.normattr = attr;
	}

//	End Member Function //

//	Member Function //

void _Cdecl termbuf::clearWindow()

//	Summary -------------------------------------------------------------
//
//		Clears the terminal window. Cursor goes to home (1,1).
//
//	End -----------------------------------------------------------------

	{
	initScreen();
	clrscr();
	gotoxy(1,1);
	gettextinfo(&ti);
	}

//	End Member Function //

//	Member Function //

void _Cdecl termbuf::gotoWindow(int x,int y)

//	Summary -------------------------------------------------------------
//
//		Moves the terminal cursor position to (x,y).
//
//	End -----------------------------------------------------------------

	{
	if(x < 1 || x > ti.winright-ti.winleft+1)	return;
	if(y < 1 || y > ti.winbottom-ti.wintop+1)	return;
	ti.curx = x;	ti.cury = y;
	initScreen();
	}

//	End Member Function //

//	Member Function //

void _Cdecl termbuf::initScreen()

//	Summary -------------------------------------------------------------
//
//		Initialize console.
//
//	Functional Description
//
//		Class termbuf maintains the screen window coordinates, current cursor
//	position, and screen attribute. Because of this, you can do other console
//	manipulation outside of the terminal window without screwing up the
//	terminal.
//
//	End -----------------------------------------------------------------

	{
	window(ti.winleft,ti.wintop,ti.winright,ti.winbottom);
	textattr(ti.attribute);
	gotoxy(ti.curx,ti.cury);
	}

//	End Member Function //

//	Member Function //

void _Cdecl termbuf::processKey(unsigned int keycode)

//	Summary -------------------------------------------------------------
//
//      Performs actions for a given keycode.
//
//	Functional Description
//
//      Member processKey() receives a keycode and performs a set of actions.
//	This member is made virtual  because you might not like the actions it
//	takes. The alternative to my approach is to have something similar to
//	the KeyMap approach - an array that correlates keycode to function
//	pointers. This is inefficient and unnecessary. If you have particular
//	requirements, just derive a new class and overload processKey(). Then
//	you can process your different requirements there, then call
//	termbuf::processKey(keycode) for remainders.
//
//	NOTE:
//
//		THIS CODE IS PRELIMINARY! I wrote this code on a laptop in the wilds
//	for christmas break from college - with no reference books. It doesn't
//	do an awful lot - just some real basics.
//
//	End -----------------------------------------------------------------

	{
	char sbuf[2];
	int ascii;

	ascii = keycode & 0xFF;

	// call putch() for printable and whitespace characters //
	if(isprint(ascii) || isspace(ascii))	putch(ascii);

	// backspace handler //
	else if(ascii == 8)						// bksp
		gotoWindow(ti.curx-1,ti.cury);

	// delete handler //
	else if(ascii == 0x7F)					// delete
		{
		movetext(ti.winleft+ti.curx,ti.wintop+ti.cury-1,
				ti.winright,ti.wintop+ti.cury-1,
				ti.winleft+ti.curx-1,ti.wintop+ti.cury-1);
		sbuf[0] = ' ';		sbuf[1] = ti.attribute;
		puttext(ti.winright,ti.wintop+ti.cury-1,
				ti.winright,ti.wintop+ti.cury-1,sbuf);
		}

	// cursor movement handlers //
	else if(keycode == 0x4700)				// home
		gotoWindow(1,1);
	else if(keycode == 0x4800)				// up arrow
		gotoWindow(ti.curx,ti.cury-1);
	else if(keycode == 0x5000)				// down
		gotoWindow(ti.curx,ti.cury+1);
	else if(keycode == 0x4B00)				// left
		gotoWindow(ti.curx-1,ti.cury);
	else if(keycode == 0x4D00)				// right
		gotoWindow(ti.curx+1,ti.cury);
	else if(keycode == 0x5100)				// pgdn
		clearWindow();
	else if(keycode == 0x4F00)				// end
		clreol();
	else if(keycode == 0x4900)				// pgup
		delline();
	}

//	End Member Function //
